/*
    C stubs for OCaml XenStore Daemon.
    Copyright (C) 2008 Patrick Colp University of British Columbia

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include <sys/mman.h>
#include <unistd.h>
#include <xenctrl.h>

#include <caml/config.h>
#include <caml/memory.h>
#include <caml/alloc.h>
#include <caml/mlvalues.h>
#include <caml/fail.h>
#include <caml/bigarray.h>

CAMLprim value
xc_domain_getinfo_c (value fd_v, value domid_v, value max_doms_v, value dominfo_v)
{
	CAMLparam4 (fd_v, domid_v, max_doms_v, dominfo_v);
	CAMLlocal1 (rv);
	
	int r, fd;
	uint32_t domid;
	unsigned int max_doms;
	void *dominfo;
	
	fd = Int_val (fd_v);
	domid = Int32_val (domid_v);
	max_doms = Int_val (max_doms_v);
	dominfo = Data_bigarray_val(dominfo_v);
	
	r = xc_domain_getinfo(fd, domid, max_doms, (xc_dominfo_t *)dominfo);
	
	rv = caml_copy_int32 (r);

	CAMLreturn (rv);
}

CAMLprim value
mb_c (value dummy)
{
	CAMLparam1 (dummy);
	
    asm volatile ( "lock; addl $0,0(%%esp)" : : : "memory" );
    
	CAMLreturn (Val_unit);
}

CAMLprim value
munmap_c (value array_v, value size_v)
{
	CAMLparam2 (array_v, size_v);
	CAMLlocal1 (rv);
	
	void *p;
	int size, r;
	
	p = Data_bigarray_val(array_v);
	size = Int_val (size_v);
	
	r = munmap(p, size);
	
	rv = caml_copy_int32 (r);
	
	CAMLreturn (rv);
}

CAMLprim value
getpagesize_c (value dummy_v)
{
	CAMLparam1 (dummy_v);
	CAMLlocal1 (rv);
	
	long size;
	size = sysconf(_SC_PAGESIZE);
	
	rv = caml_copy_int32 (size);
	
	CAMLreturn (rv);
}

CAMLprim value
mmap_c (value fd_v, value pagesize_v)
{
	CAMLparam2 (fd_v, pagesize_v);
	CAMLlocal1 (rv);
	
	int fd, pagesize;
	void *r;
	long dims[1];
	
	fd = Int_val (fd_v);
	pagesize = Int_val (pagesize_v);
	
	r = mmap(NULL, pagesize, PROT_READ|PROT_WRITE, MAP_SHARED, fd, 0);
	dims[0] = pagesize;
	
	rv = alloc_bigarray(BIGARRAY_UINT8 | BIGARRAY_C_LAYOUT, 1, r, dims);
//	rv = alloc_bigarray(BIGARRAY_INT32 | BIGARRAY_C_LAYOUT, 1, r, pagesize);
	
	CAMLreturn (rv);
}

CAMLprim value
xc_map_foreign_range_c (value xc_handle_v, value domid_v, value pagesize_v, value mfn_v)
{
	CAMLparam4 (xc_handle_v, domid_v, pagesize_v, mfn_v);
	CAMLlocal1 (rv);
	
	int xc_handle, pagesize;
	uint32_t domid;
	unsigned long mfn;
	void *r;
	long dims[1];
	
	xc_handle = Int_val (xc_handle_v);
	domid = Int_val (domid_v);
	pagesize = Int32_val (pagesize_v);
	mfn = Int_val (mfn_v);
	
	r = xc_map_foreign_range(xc_handle, domid, pagesize, PROT_READ|PROT_WRITE, mfn);
	dims[0] = pagesize;
	
	rv = alloc_bigarray(BIGARRAY_UINT8 | BIGARRAY_C_LAYOUT, 1, r, dims);
	
	CAMLreturn (rv);
}

CAMLprim value
xc_evtchn_bind_interdomain_c (value xce_handle_v, value domid_v, value remote_port_v)
{
    CAMLparam3 (xce_handle_v, domid_v, remote_port_v);
    CAMLlocal1 (rv);
    
    int xce_handle, domid, r;
    uint32_t remote_port;
    
    xce_handle = Int_val (xce_handle_v);
    domid = Int_val (domid_v);
    remote_port = Int32_val (remote_port_v);
    
    r = xc_evtchn_bind_interdomain (xce_handle, domid, remote_port);
    
    rv = caml_copy_int32 (r);
    
    CAMLreturn (rv);
}

CAMLprim value
xc_evtchn_bind_virq_c (value xce_handle_v, value virq_v)
{
    CAMLparam2 (xce_handle_v, virq_v);
    CAMLlocal1 (rv);
    
    int xce_handle, r;
    unsigned int virq;
    
    xce_handle = Int_val (xce_handle_v);
    virq = Int_val (virq_v);
    
    r = xc_evtchn_bind_virq (xce_handle, virq);
    
    rv = caml_copy_int32 (r);
    
    CAMLreturn (rv);
}

CAMLprim value
xc_evtchn_fd_c (value xce_handle_v)
{
    CAMLparam1 (xce_handle_v);
    CAMLlocal1 (rv);
    
    int xce_handle, r;
    
    xce_handle = Int_val (xce_handle_v);
    
    r = xc_evtchn_fd (xce_handle);
    
    rv = caml_copy_int32 (r);
    
    CAMLreturn (rv);
}

CAMLprim value
xc_evtchn_notify_c (value xce_handle_v, value port_v)
{
    CAMLparam2 (xce_handle_v, port_v);
    CAMLlocal1 (rv);
    
    int xce_handle, r;
    uint32_t port;
    
    xce_handle = Int_val (xce_handle_v);
    port = Int32_val (port_v);
    
    r = xc_evtchn_notify (xce_handle, port);
    
    rv = caml_copy_int32 (r);
    
    CAMLreturn (rv);
}

CAMLprim value
xc_evtchn_open_c (value unit)
{
    CAMLparam1 (unit);
    CAMLlocal1 (rv);
    
    int r;
    
    r = xc_evtchn_open ();
    
    rv = caml_copy_int32 (r);
    
    CAMLreturn (rv);
}

CAMLprim value
xc_evtchn_pending_c (value xce_handle_v)
{
    CAMLparam1 (xce_handle_v);
    CAMLlocal1 (rv);
    
    int xce_handle, r;
    
    xce_handle = Int_val (xce_handle_v);
    
    r = xc_evtchn_pending (xce_handle);
    
    rv = caml_copy_int32 (r);
    
    CAMLreturn (rv);
}

CAMLprim value
xc_evtchn_unbind_c (value xce_handle_v, value port_v)
{
    CAMLparam2 (xce_handle_v, port_v);
    CAMLlocal1 (rv);
    
    int xce_handle, r;
    uint32_t port;
    
    xce_handle = Int_val (xce_handle_v);
    port = Int32_val (port_v);
    
    r = xc_evtchn_unbind (xce_handle, port);
    
    rv = caml_copy_int32 (r);
    
    CAMLreturn (rv);
}

CAMLprim value
xc_evtchn_unmask_c (value xce_handle_v, value port_v)
{
    CAMLparam2 (xce_handle_v, port_v);
    CAMLlocal1 (rv);
    
    int xce_handle, r;
    uint32_t port;
    
    xce_handle = Int_val (xce_handle_v);
    port = Int32_val (port_v);
    
    r = xc_evtchn_unmask (xce_handle, port);
    
    rv = caml_copy_int32 (r);
    
    CAMLreturn (rv);
}

CAMLprim value
xc_interface_open_c (value unit)
{
    CAMLparam1 (unit);
    CAMLlocal1 (rv);
    
    int r;
    
    r = xc_interface_open ();
    
    rv = caml_copy_int32 (r);
    
    CAMLreturn (rv);
}

CAMLprim value
xc_interface_close_c (value xc_handle_v)
{
    CAMLparam1 (xc_handle_v);
    CAMLlocal1 (rv);
    
    int xc_handle, r;
    
    xc_handle = Int_val (xc_handle_v);
    
    r = xc_interface_close (xc_handle);
    
    rv = caml_copy_int32 (r);
    
    CAMLreturn (rv);
}
